#!/usr/bin/env python
#
# This script can be used to create an initial conda recipe repository
# for a new FSL project, or update an existing recipe repository.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import os.path  as op
import textwrap as tw
import             os
import             sys
import             glob
import             shutil
import             argparse
from unittest import mock

import jinja2   as j2

from fsl_ci_utils            import (USERNAME,
                                     EMAIL,
                                     indir,
                                     tempdir,
                                     sprun)
from fsl_ci_utils.gitlab_api import (project_exists,
                                     list_project_branches,
                                     create_repository,
                                     get_project_version,
                                     open_merge_request)
from fsl_ci_utils.conda      import  gen_recipe_path

import fsl_project_dependencies as fsldeps


SERVER_URL = 'https://git.fmrib.ox.ac.uk'
"""Default gitlab instance URL, if not specified on the command.line. """


TEMPLATE_DIR = op.abspath(op.join(op.dirname(__file__), 'templates'))
"""Directory which contains conda recipe template files. """


MERGE_REQUEST_MSG = """
Update conda recipe to latest template-generated recipe. This merge request
was automatically generated by fsl-ci-rules/utils/create_conda_recipe.py.
""".strip()


EXCLUDE = [
    'boost-cpp',
    'sgeutils',
    'glibmm',
    'libxmlpp',
    'libxml2',
    'fslio',
]


VERSION_OVERRIDE = {
    'libgd'     : '2.2.5',
    'libnlopt'  : '2.6.2',
    'libpng'    : '1.6.37',
    'libsqlite' : '3.33.0',
    'tk'        : '8.6.10',
    'zlib'      : '1.2.11',
    'ciftilib'  : '1.5.3',
}
"""Some hard-coded version numbers for external ("extras") FSL dependencies.
"""


def project_type(project_dir):
    """Return:
      - "cpp" if the project looks like a conventional C/C++ Makefile-based
        project
      - "python" if the project looks like a Python based project
      - "other" if the project looks like a non-C/C++ Makefile based project
      - "unwknown" if the project doesn't look like anything we can work with.
    """

    if op.exists(op.join(project_dir, 'setup.py')):
        return 'python'

    elif op.exists(op.join(project_dir, 'Makefile')):
        for pattern in ['*.cc', '*.h', '*.cpp', '*.cxx', '*.c']:
            if glob.glob(op.join(project_dir, '**', pattern), recursive=True):
                return 'cpp'
        return 'other'
    else:
        return 'unknown'


def gen_repository_url(project_path, server, token=None):
    """Generates a URL for the given project. """

    # we've been given a full URL to a git repo
    if project_path.startswith('http') or project_path.startswith('git'):
        return project_path

    if token is not None:
        prefix, suffix = server.split('://')
        server         = f'{prefix}://gitlab-ci-token:{token}@{suffix}'

    return f'{server}/{project_path}.git'


def get_version(project_path, server, token):
    """Return the most recent tag on the given project, or '' if the project
    is not a gitlab project path.
    """
    if project_path.startswith('http') or project_path.startswith('git'):
        return ''
    else:
        return get_project_version(project_path, server, token)


def clone_repository(path, server, token, destdir, branch='master'):
    """Clone the specified gitlab repository into destdir. """
    url = gen_repository_url(path, server, token)
    sprun(f'git clone {url} {destdir}')
    sprun(f'git checkout {branch}', cwd=destdir)


def create_local_recipe_repository(recipe_dir,
                                   recipe_path,
                                   server,
                                   token):
    """If recipe_path already exists on gitlab, it is cloned into recipe_dir.
    Otherwise recipe_dir is turned into a git repository with an initial
    commit and remote origin configured.
    """
    recipe_url = gen_repository_url(recipe_path, server, token)

    os.mkdir(recipe_dir)
    with indir(recipe_dir):
        sprun( 'git init .')
        sprun(f'git remote add origin {recipe_url}')


def commit_and_push_recipe_repository(recipe_dir,
                                      project_path,
                                      branch='master',
                                      init=True,
                                      push=True):
    """Commit the current state of the local recipe repository and push
    it to gitlab.
    """
    with indir(recipe_dir):

        if init:
            msg = f'MNT: Initial auto-generated conda ' \
                  f'recipe for {project_path}'
        else:
            msg = f'MNT: Re-generated conda recipe for {project_path}'

        sprun(f'git config user.name {USERNAME}')
        sprun(f'git config user.email {EMAIL}')
        sprun(f'git checkout -B {branch}')
        sprun( 'git add *')
        sprun(f'git commit -m "{msg}"')

        if push:
            sprun(f'git push origin {branch}')


def get_fsl_project_dependencies(project_path, project_dir, server, token):
    """Returns a list of all FSL dependencies of the project as (project,
    version) tuples.
    """
    if ('FSLDIR' not in os.environ) or ('FSLDEVDIR' not in os.environ):
        print('FSLDIR/FSLDEVDIR are not set - cannot identify '
              f'dependencies for project {project_path}')
        return []

    deps = fsldeps.get_project_dependencies(project_dir)
    deps = [d.lower() for d in deps if d not in EXCLUDE]
    vers = []


    for i, dep in enumerate(deps):
        if dep in VERSION_OVERRIDE:
            dver = VERSION_OVERRIDE[dep]
        else:
            try:
                # assume that the dependency has
                # a repo at "server/fsl/dep"
                dver = get_project_version(f'fsl/{dep}', server, token)
                dep  = gen_recipe_path(dep).rsplit('/')[-1]

            except Exception:
                dver = 'UNKNOWN'

        vers.append(dver)
        deps[i] = dep

    deps = list(zip(deps, vers))
    deps = sorted([d for d in deps if     d[0].startswith('fsl')]) + \
           sorted([d for d in deps if not d[0].startswith('fsl')])
    return deps


def get_setup_py_metadata(project_dir):
    """Attempts to extract project metadata from its setup.py file. """
    setup_py = op.abspath(op.join(project_dir, 'setup.py'))

    if not op.exists(setup_py):
        return {}

    setup_meta = {}

    def setup(**kwargs):
        setup_meta.update(kwargs)

    with open(setup_py) as f:
        code = f.read()

    ns = {
        '__name__'   : '__main__',
        '__doc__'    : None,
        '__file__'   : setup_py,
    }

    with mock.patch('setuptools.setup',           setup),                   \
         mock.patch('distutils.core.setup',       setup),                   \
         mock.patch('numpy.distutils.core.setup', setup),                   \
         mock.patch.dict('sys.modules', {'versioneer' : mock.MagicMock()}), \
         indir(project_dir):

        try:
            code = compile(code, setup_py, 'exec', dont_inherit=True)
            exec(code, ns, ns)
        except Exception as e:
            print(f'Could not extract project metadata from {setup_py}: {e}')

    return setup_meta


def get_python_entrypoints(project_dir):
    """Return entry points provided by a python project managed by the
    given setup.py file. Will only work for simple setup.py files which
    don't try to do anything too complicated.

    Takes inspiration from conda_build._load_setup_py_data.
    """

    setup_meta = get_setup_py_metadata(project_dir)

    if 'entry_points' in setup_meta:
        entry_points = setup_meta['entry_points']
        entry_points = entry_points.get('console_scripts', []) + \
                       entry_points.get('gui_scripts',     [])
    else:
        entry_points = []

    return sorted(entry_points)


def get_python_executables(project_dir):
    """Return a list of all executables provided by the project."""

    setup_meta  = get_setup_py_metadata(project_dir)
    entrypoints = get_python_entrypoints(project_dir)
    entrypoints = [ep.split('=')[0].strip() for ep in entrypoints]

    scripts     = setup_meta.get('scripts', [])
    scripts     = [op.basename(s) for s in scripts]

    return entrypoints + scripts


def create_python_recipe_template(project_path,
                                  recipe_path,
                                  project_dir,
                                  recipe_dir,
                                  server,
                                  token):
    """Creates a conda recipe for the Python project in project_dir, storing
    the recipe files in recipe_dir.
    """
    recipe_name        = recipe_path.rsplit('/')[-1]
    metayaml_template  = op.join(TEMPLATE_DIR, 'meta.yaml.python.template')
    postlink_template  = op.join(TEMPLATE_DIR, 'post-link.sh.template')
    preunlink_template = op.join(TEMPLATE_DIR, 'pre-unlink.sh.template')
    metayaml_dest      = op.join(recipe_dir,   'meta.yaml')
    postlink_dest      = op.join(recipe_dir,   'post-link.sh')
    preunlink_dest     = op.join(recipe_dir,   'pre-unlink.sh')

    base_version       = get_project_version('fsl/base',   server, token)
    project_version    = get_version(        project_path, server, token)
    entrypoints        = get_python_entrypoints(project_dir)
    exes               = get_python_executables(project_dir)

    meta               = {
        'template_name'              : recipe_name,
        'template_repository'        : gen_repository_url(project_path,
                                                          server),
        'template_recipe_repository' : gen_repository_url(recipe_path,
                                                          server),
        'template_version'           : project_version,
        'fslbaseversion'             : base_version,
        'entrypoints'                : entrypoints,
        'exes'                       : exes,
    }

    # post-link/pre-unlink only needed for
    # projects which provide executables
    templates = [(metayaml_template, metayaml_dest)]
    if len(exes) > 0:
        templates.append((postlink_template,  postlink_dest))
        templates.append((preunlink_template, preunlink_dest))

    for template, dest in templates:
        with open(template, 'rt') as f:
            loader   = j2.FileSystemLoader(TEMPLATE_DIR)
            env      = j2.Environment(loader=loader)
            template = env.from_string(f.read())
        with open(dest, 'wt') as f:
            f.write(template.render(**meta) + '\n')


def create_cpp_recipe_template(project_path,
                               recipe_path,
                               project_dir,
                               recipe_dir,
                               server,
                               token,
                               metayaml_template='meta.yaml.cpp.template'):
    """Creates a conda recipe for the C/C++ project in project_dir, storing the
    recipe files in recipe_dir.
    """

    recipe_name        = recipe_path.rsplit('/')[-1]
    buildsh_template   = op.join(TEMPLATE_DIR, 'build.sh.template')
    metayaml_template  = op.join(TEMPLATE_DIR, metayaml_template)
    postlink_template  = op.join(TEMPLATE_DIR, 'post-link.sh.template')
    preunlink_template = op.join(TEMPLATE_DIR, 'pre-unlink.sh.template')
    buildsh_dest       = op.join(recipe_dir,   'build.sh')
    metayaml_dest      = op.join(recipe_dir,   'meta.yaml')
    postlink_dest      = op.join(recipe_dir,   'post-link.sh')
    preunlink_dest     = op.join(recipe_dir,   'pre-unlink.sh')
    base_version       = get_project_version('fsl/base',   server, token)
    project_version    = get_version(        project_path, server, token)
    exes               = fsldeps.get_project_executables(project_dir)
    deps               = get_fsl_project_dependencies(project_path,
                                                      project_dir,
                                                      server,
                                                      token)
    meta = {
        'template_name'              : recipe_name,
        'template_repository'        : gen_repository_url(project_path, server),
        'template_recipe_repository' : gen_repository_url(recipe_path, server),
        'template_version'           : project_version,
        'fslbaseversion'             : base_version,
        'dependencies'               : deps,
        'exes'                       : exes,
    }

    with open(buildsh_template, 'rt') as inf, \
         open(buildsh_dest, 'wt') as outf:
        outf.write(inf.read())

    # post-link/pre-unlink only needed for
    # projects which provide executables
    templates = [(metayaml_template, metayaml_dest)]
    if len(exes) > 0:
        templates.append((postlink_template,  postlink_dest))
        templates.append((preunlink_template, preunlink_dest))

    for template, dest in templates:
        with open(template, 'rt') as f:
            loader   = j2.FileSystemLoader(TEMPLATE_DIR)
            env      = j2.Environment(loader=loader)
            template = env.from_string(f.read())
        with open(dest, 'wt') as f:
            f.write(template.render(**meta) + '\n')


def create_other_recipe_template(project_path,
                                 recipe_path,
                                 project_dir,
                                 recipe_dir,
                                 server,
                                 token):
    """Creates a conda recipe for the FSL project in project_dir, storing the
    recipe files in recipe_dir.
    """
    create_cpp_recipe_template(project_path,
                               recipe_path,
                               project_dir,
                               recipe_dir,
                               server,
                               token,
                               'meta.yaml.other.template')


def create_recipe_template(project_path,
                           recipe_path,
                           project_dir,
                           recipe_dir,
                           server,
                           token):
    """Creates a conda recipe template for the given project. """

    ptype = project_type(project_dir)
    if ptype == 'cpp':
        create_cpp_recipe_template(project_path,
                                   recipe_path,
                                   project_dir,
                                   recipe_dir,
                                   server,
                                   token)
    elif ptype == 'python':
        create_python_recipe_template(project_path,
                                      recipe_path,
                                      project_dir,
                                      recipe_dir,
                                      server,
                                      token)
    elif ptype == 'other':
        create_other_recipe_template(project_path,
                                     recipe_path,
                                     project_dir,
                                     recipe_dir,
                                     server,
                                     token)
    else:
        print(f'{project_path} does not look like '
              'a FSL or Python project - aborting.')
        sys.exit(1)


def gen_branch_name(recipe_path, server, token):
    """Generates a unique branch name on an existing recipe repository
    to open a MR with the updated recipe.
    """

    branch      = 'mnt/recipe_update'
    allbranches = list_project_branches(recipe_path, server, token)

    count = 0
    while branch in allbranches:
        count += 1
        branch = f'mnt/recipe_update-{count}'

    return branch


def parseArgs():
    usage  = 'Usage: create_conda_recipe.py -t <token> [options] project_path'
    desc   = tw.dedent("""
    Create or update a conda recipe repository for the specified project.
    """).strip()
    parser = argparse.ArgumentParser(
        usage=usage,
        description=desc,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    helps = {
        'project_path' :
        'Path (namespace/project) of the gitlab project repository, or URL '
        'to remote (e.g. github) repository.',

        'recipe_path' :
        'Path (namespace/project) of the gitlab recipe repository (default: '
        'automatically generated from project path). ',

        'server' :
        f'Gitlab server (default: {SERVER_URL})',

        'token' :
        'Gitlab API access token with read+write access',

        'branch' :
        'Checkout this branch from the project repository (default: master)',

        'local' :
        'Do not push recipe repository to gitlab - instead, create a local '
        'repository at the specified path.',

        'nopush' :
        'Do not push a branch, or open a merge request, on the remote recipe '
        'repository (if one exists)',
    }

    parser.add_argument('project_path',        help=helps['project_path'])
    parser.add_argument('-r', '--recipe_path', help=helps['recipe_path'])
    parser.add_argument('-l', '--local',       help=helps['local'])
    parser.add_argument('-n', '--nopush',      help=helps['nopush'],
                        action='store_true')
    parser.add_argument('-b', '--branch',      help=helps['branch'],
                        default='master')
    parser.add_argument('-s', '--server',      help=helps['server'],
                        default=SERVER_URL)
    parser.add_argument('-t', '--token',       help=helps['token'],
                        required=True)

    args = parser.parse_args()

    args.token = args.token.strip()

    if args.local is not None:
        args.local = op.abspath(args.local)
    if args.recipe_path is None:
        args.recipe_path = gen_recipe_path(args.project_path)

    return args


def main():
    args = parseArgs()

    with tempdir():
        project_dir   = op.abspath('project')
        recipe_dir    = op.abspath('recipe')
        recipe_exists = project_exists(args.recipe_path,
                                       args.server,
                                       args.token)

        print(f'Cloning {args.project_path}')
        clone_repository(args.project_path,
                         args.server,
                         args.token,
                         project_dir,
                         args.branch)

        if recipe_exists:
            print(f'Cloning existing recipe repository ({args.recipe_path})')
            clone_repository(args.recipe_path,
                             args.server,
                             args.token,
                             recipe_dir)
        else:
            print('Creating local recipe repository')
            create_local_recipe_repository(recipe_dir,
                                           args.recipe_path,
                                           args.server,
                                           args.token)

        print('(Re-)Generating conda recipe')
        create_recipe_template(args.project_path,
                               args.recipe_path,
                               project_dir,
                               recipe_dir,
                               args.server,
                               args.token)

        if recipe_exists:
            branch = gen_branch_name(args.recipe_path,
                                     args.server,
                                     args.token)
        else:
            branch = 'master'

        # create remote repository on gitlab
        if (args.local is None) and (not recipe_exists):
            print(f'Creating gitlab repository ({args.recipe_path})')
            create_repository(args.recipe_path,
                              args.server,
                              args.token)

        print('Commiting recipe and (maybe) pushing')
        commit_and_push_recipe_repository(recipe_dir,
                                          args.recipe_path,
                                          branch=branch,
                                          init=(not recipe_exists),
                                          push=(not args.nopush))

        if args.local is not None:
            print(f'Copying repository to {args.local}')
            shutil.copytree(recipe_dir, args.local)

        # open MR
        if recipe_exists and (not args.nopush):
            print('Opening MR on existing recipe '
                  f'repository ({args.recipe_path})')
            open_merge_request(args.recipe_path,
                               branch,
                               MERGE_REQUEST_MSG,
                               args.server,
                               args.token)


if __name__ == '__main__':
    main()
