#!/usr/bin/env python
#
# Convenience script which checks the package status of all FSL conda recipes.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This script generates a table containing information about every conda
recipe in the gitlab fsl/conda/ space. The table contains columns for:
 - The recipe package name
 - The corresponding project repository
 - The latest tag on the project repository
 - The version listed in the recipe meta.yaml file
 - The latest package version in the FSL production channel
 - The latest package version in the FSL staging channel
 - (during development) The number of commits that the project mnt/conda
   branch is ahead/behind the master branch.
"""


import sys
import argparse
import itertools as it
import urllib.parse as urlparse

from fsl_ci_utils.gitlab_api import (get_projects_in_namespace,
                                     download_file,
                                     lookup_project_tags,
                                     get_divergence)
from fsl_ci_utils.conda      import (Package,
                                     load_meta_yaml,
                                     get_channel_packages,
                                     load_packages_from_recipe_repositories,
                                     build_dependency_graph)


SERVER_URL = 'https://git.fmrib.ox.ac.uk'
"""Default gitlab instance URL, if not specified on the command.line."""


RECIPE_NAMESPACE = "fsl/conda"
"""Default recipe namespace, if not specified on the command.line."""


STAGING_CHANNEL_URL = \
    "http://f162.lap.fmrib.ox.ac.uk:8888/test_conda_channels/staging"
"""Default staging channel URL, if not specified on the command.line."""


PRODUCTION_CHANNEL_URL = \
    "http://f162.lap.fmrib.ox.ac.uk:8888/test_conda_channels/production"
"""Default production channel URL, if not specified on the command.line."""


MISSING = '-'
"""Value to use in output table for missing/irrelevant cells. """


def get_recipe_info(project_path, server, token):
    """Return the name, version, and project repository of the given conda
    recipe repository.
    """
    def getval(d, *keys, default=None):
        for k in keys:
            d = d.get(k, None)
            if d is None:
                break

        if d is None: return default
        else:         return d

    metayaml   = download_file(project_path, 'meta.yaml', server, token)
    metayaml   = load_meta_yaml(metayaml)
    name       = getval(metayaml, 'package', 'name',    default=MISSING)
    version    = getval(metayaml, 'package', 'version', default=MISSING)
    repository = getval(metayaml, 'source',  'git_url', default=MISSING)
    noarch     = getval(metayaml, 'build',   'noarch',  default=MISSING)

    if noarch in ('generic', 'python'):
        arch = noarch
    else:
        reqs = getval(metayaml, 'requirements', 'host', default=[])
        if 'compiler' in reqs:
            arch = 'native'
        else:
            arch = 'generic (needs update)'

    return {
        'name'       : name,
        'version'    : version,
        'repository' : repository,
        'arch'       : arch,
    }


def get_project_info(project_repo, server, token):
    """Return the project name, and most recent tag, for the given
    project repository.
    """

    name = MISSING
    tag  = MISSING

    if project_repo is not None:
        try:
            project_path = urlparse.urlparse(project_repo)[2][1:]
            if project_path.endswith('.git'):
                project_path = project_path[:-4]
            name         = project_path.split('/')[-1]
            tags         = lookup_project_tags(project_path, server, token)

            if len(tags) > 0: tag = tags[0]
            else:             tag = MISSING
        # the project repo is probably not on gitlab (e.g. on github)
        except Exception as e:
            pass

    return {
        'repository' : project_repo,
        'path'       : project_path,
        'name'       : name,
        'tag'        : tag
    }


def sort_projects(projects, server, token):
    """Sorts projects according to their position in the dependency hierarchy.
    Returned as a list of lists, where each child list contains a set of
    packages that do not depend on each other, and thus can be built in
    parallel.
    """

    def get_next_build_group(pkgs, previous):
        group = []
        for name, pkg in pkgs.items():
            reqs = pkg.dependencies
            reqs = {r.name for r in reqs if isinstance(r, Package)}
            if len(reqs.difference(previous)) == 0 and \
               (name not in previous)              and \
               (name not in group):
                group.append(name)
        return group

    def get_build_group_order(pkgs):
        groups   = []
        previous = []
        while len(previous) != len(pkgs):
            group = get_next_build_group(pkgs, previous)
            groups  .append(group)
            previous.extend(group)
        return groups

    pkgs     = load_packages_from_recipe_repositories(projects, server, token)
    groups   = get_build_group_order(pkgs)
    projects = []

    for group in groups:
        projects.append([])
        for pkgname in group:
            projects[-1].append(pkgs[pkgname].recipe_path)

    return projects


def parseArgs():

    helps = {
        'token' :
        'Gitlab API access token with read+write access',

        'server' :
        f'Gitlab server (default: {SERVER_URL})',

        'namespace' :
        'Gitlab namespace which contains the conda '
        f'recipes (default: {RECIPE_NAMESPACE}',

        'staging_url'    : 'URL of the staging conda channel',
        'production_url' : 'URL of the production conda channel',

        'output' :
        'Save to this output file, instead of printing to standard output',

        'exclude' :
        'Exclude packages with this repository name. Can be used '
        'multiple times. The "docs" project is always excluded.'
    }

    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--server', default=SERVER_URL,
                        help=helps['server'])
    parser.add_argument('-t', '--token', required=True,
                        help=helps['token'])
    parser.add_argument('-n', '--namespace',
                        help=helps['namespace'],
                        default=RECIPE_NAMESPACE)
    parser.add_argument('-su', '--staging_url',
                        help=helps['staging_url'],
                        default=STAGING_CHANNEL_URL)
    parser.add_argument('-pu', '--production_url',
                        help=helps['production_url'],
                        default=PRODUCTION_CHANNEL_URL)
    parser.add_argument('-o', '--output',
                        help=helps['output'])
    parser.add_argument('-e', '--exclude',
                        help=helps['exclude'], action='append')

    args = parser.parse_args()

    if args.exclude is None:
        args.exclude = []

    args.exclude.append('docs')

    return args


def print_columns(titles, columns, f=None):
    """Convenience function which pretty-prints a collection of columns in a
    tabular format.

    :arg titles:  A list of titles, one for each column.

    :arg columns: A list of columns, where each column is a list of strings.
    """

    if f is None:
        f = sys.stdout

    cols  = []

    for t, c in zip(titles, columns):
        cols.append([t] + list(map(str, c)))

    columns = cols
    colLens = []

    for col in columns:
        maxLen = max([len(r) for r in col])
        colLens.append(maxLen)

    fmtStr = ' | '.join(['{{:<{}s}}'.format(l) for l in colLens]) + '\n'

    titles  = [col[0]  for col in columns]
    columns = [col[1:] for col in columns]

    separator = ['-' * l for l in colLens]

    f.write(fmtStr.format(*titles))
    f.write(fmtStr.format(*separator))

    nrows = len(columns[0])
    for i in range(nrows):

        row = [col[i] for col in columns]
        f.write(fmtStr.format(*row))


def filter_projects(projects, exclude):
    """Removes projects which the user has requested be excluded.
    """

    if exclude is None:
        return projects

    filtered = []
    for project in projects:

        name = project.rsplit('/')[-1]

        if name in exclude:
            print(f'Removing {project}')
        else:
            filtered.append(project)

    return filtered


def get_mnt_conda_divergence(project, server, token):
    """Get the number of commits ahead/behind master the
    mnt/conda branch is, or 'n/a', 'n/a' if not relevant.
    """

    repo = project['repository']
    path = project['path']

    if server not in repo:
        return MISSING, MISSING

    ca, cb = get_divergence(path,
                            server,
                            token, 'master', 'mnt/conda')

    if ca is None: ca = MISSING
    else:          ca = len(ca)
    if cb is None: cb = MISSING
    else:          cb = len(cb)

    return ca, cb


def main():
    args       = parseArgs()
    projects   = get_projects_in_namespace(args.namespace,
                                           args.server,
                                           args.token)

    projects   = filter_projects(projects, args.exclude)
    projects   = sort_projects(projects, args.server, args.token)
    staging    = get_channel_packages(args.staging_url)
    production = get_channel_packages(args.production_url)

    # Print a table containing the following columns
    # package name, recipe version, project repo <- get_recipe_info
    # latest project tag                         <- get_project_info
    # staging version, production version        <- get_channel_version
    # mnt/conda commits ahead/behind master      <- get_mnt_conda_divergence
    titles = ['Package', 'Repository', 'Recipe version',
              'Latest tag', 'Staging', 'Production',
              'Ahead', 'Behind']
    packagenames = []
    recipevers   = []
    projectrepos = []
    projecttags  = []
    stagingvers  = []
    prodvers     = []
    ahead        = []
    behind       = []

    # add a divider between each group
    projects = list(it.chain(*[group + [None] for group in projects]))

    for project_path in projects:

        if project_path is None:
            packagenames.append('')
            recipevers  .append('')
            projectrepos.append('')
            projecttags .append('')
            stagingvers .append('')
            prodvers    .append('')
            ahead       .append('')
            behind      .append('')
            continue

        recipe  = get_recipe_info(project_path, args.server, args.token)
        repo    = recipe['repository']
        package = recipe['name']
        project = get_project_info(repo, args.server, args.token)
        ca, cb  = get_mnt_conda_divergence(project, args.server, args.token)

        if package in staging:    stagingver = staging[package].version
        else:                     stagingver = MISSING
        if package in production: prodver    = production[package].version
        else:                     prodver    = MISSING

        if args.server in repo:
            repo = repo.replace(args.server, '<git@fmrib>')

        packagenames.append(package)
        recipevers  .append(recipe['version'])
        projectrepos.append(repo)
        projecttags .append(project['tag'])
        stagingvers .append(stagingver)
        prodvers    .append(prodver)
        ahead       .append(ca)
        behind      .append(cb)

    if args.output is not None:
        outf = open(args.output, 'wt')
    else:
        outf = None

    try:
        print_columns(titles,
                      (packagenames, projectrepos, recipevers,
                       projecttags, stagingvers, prodvers,
                       ahead, behind),
                      outf)
    finally:
        if outf is not None:
            outf.close()


if __name__ == '__main__':
    main()
