#!/usr/bin/env python
#
# trigger_build.py - Trigger a package build and deployment on one or more FSL
# conda recipe repositories.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import os.path               as op
import functools             as ft
import textwrap              as tw
import                          argparse
import multiprocessing.dummy as mp


import fsl_ci_utils.gitlab_api as      gitlab
from   fsl_ci_utils.gitlab_api import (trigger_job,
                                       find_latest_job,
                                       trigger_pipeline,
                                       wait_on_pipeline)


gitlab.VERBOSE = False


SERVER_URL = 'https://git.fmrib.ox.ac.uk'
"""Default gitlab instance URL, if not specified on the command.line."""


def trigger_build(project, server, token, production):
    """Triggers a pipeline on the master branch of project and waits for it
    to complete.
    """

    if production:
        channel   = 'production'
        variables = {}
    else:
        channel   = 'staging'
        variables = {'STAGING' : 'true'}

    try:
        pipeline = trigger_pipeline(project, 'master', server, token, variables)
    except:
        return None

    pid = pipeline['id']

    print(f'Pipeline triggered on {project} ({channel} build) '
          f'- see {pipeline["web_url"]}')

    try:
        status = wait_on_pipeline(project, pid, server, token)
    except:
        return None

    print(f'Build pipeline for {project} has finished: {status}')

    if status != 'manual':
        return None

    return pid


def trigger_deploy(project, pid, server, token, production):
    """Triggers the most recently created manual 'deploy-conda-package' job,
    and waits for it to complete.
    """

    # deployment to staging/production gets set at
    # build time, so we don't need to pass the STAGING
    # variable here, like we do in trigger_build
    if production: channel = 'production'
    else:          channel = 'staging'

    print(f'Triggering deploy-conda-package job '
          f'on {project} (deploying to {channel} channel)')

    try:
        jid = find_latest_job(project, server, token,
                              'deploy-conda-package', 1)['id']
        trigger_job(project, jid, server, token)

        status = wait_on_pipeline(project, pid, server, token)
        print(f'Deploy job {project} has finished: {status}')
    except Exception as e:
        print(f'Error triggering deploy job on {project}: {e}')


def parseArgs():
    """Parses and returns command line arguments. """

    name   = op.basename(__file__)
    usage  = f'Usage: {name} -t <token> [options] project [project ...]'
    desc  = tw.dedent("""
    Trigger a package build and deployment on
    one or more FSL conda recipe repositories.
    """).strip()

    helps = {
        'token'      : 'Gitlab API access token with read+write access',
        'server'     :  f'Gitlab server (default: {SERVER_URL})',
        'project'    : 'Project(s) to build',
        'production' : 'Deploy built package(s) to production cnannel '
                       '(default: deploy to staging channel)'
    }

    parser = argparse.ArgumentParser(usage=usage, description=desc)
    parser.add_argument('project', nargs='+',
                        help=helps['project'])
    parser.add_argument('-s', '--server', default=SERVER_URL,
                        help=helps['server'])
    parser.add_argument('-t', '--token', required=True,
                        help=helps['token'])
    parser.add_argument('-p', '--production', action='store_true',
                        help=helps['production'])

    return parser.parse_args()


def main():
    """Trigger builds on all listed projects concurrently, and wait for them
    all to complete or fail.
    """

    args     = parseArgs()
    projects = args.project
    pool     = mp.Pool(len(projects))

    build  = ft.partial(trigger_build,
                        server=args.server,
                        token=args.token,
                        production=args.production)
    deploy = ft.partial(trigger_deploy,
                        server=args.server,
                        token=args.token,
                        production=args.production)

    # Build packages in parallel
    pids = pool.map(build, projects)
    pool.close()
    pool.join()

    # Deploy sequentially, to avoid conflicts
    # arising from simultaneous access to the
    # channel directory
    for project, pid in zip(projects, pids):
        if pid is not None:
            deploy(project, pid)


if __name__  == '__main__':
    main()
